# _*_ coding: utf-8 _*_
#
# hspylib v0.11.155
#
# Package: main.hspylib.modules.fetch
"""Package initialization."""

__all__ = [
  'fetch', 
  'http_response', 
  'uri_builder', 
  'uri_scheme'
]
__version__ = '0.11.155'
